/*
 * Copyright (c) 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <OS/math.h>

extern "C" {
#include <projects.h>
int set_settings(float startaz, float endaz,
		  float obsx, float obsy, float obsheight, /* float obswidth,*/
		  float tarx, float tary, float tarheight, float tarwidth,
		  float range, long vertacc,
		  int priormask, /* unused? */
		  int prodtype, /* enum { VIS, VIP } 2 or 4 valued */
		  int prof /* viewshed or profile */);
void do_los();
}

#include <LosServ/los.h>
#include <LosServ/losfunc.h>
#include <Dem/usgsdem.h>
#include <ComTerp/comhandler.h>
#include <ComTerp/comterpserv.h>
#include <ComTerp/comvalue.h>
#include <Map/mapfclass.h>
#include <Map/mapfeature.h>
#include <Map/mapprim.h>
#include <MapServ/mapfunc.h>
#include <MapUnidraw/mapcomps.h>
#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>
#include <TopoFace/fgeomobjs.h>
#include <OS/memory.h>
#ifndef MAKEDEPEND
#ifdef VTK
#include <vtkTriangle.h>
#endif
#endif

#ifdef HAVE_ACE
#include <ace/SOCK_Connector.h>
#endif

#include <fstream.h>
#include <strstream.h>
#include <stream.h>
#include <stdio.h>

#if defined(__svr4__) || defined(__linux__)
#include <values.h>
#endif

int round(float x) { return x > 0 ? int(x+0.5) : -int(-x+0.5); }
//int round(double x) { return x > 0 ? int(x+0.5) : -int(-x+0.5); }
UsgsDem* dem;

/*****************************************************************************/

DemLosFunc::DemLosFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void DemLosFunc::execute() {
    ComValue lostype = pop_stack(); /* 0=viewshed, 1=profile */
    ComValue valtype = pop_stack(); /* 0 = 2val, 1 = 4val */
    ComValue range = pop_stack();
    ComValue targwid = pop_stack();
    ComValue targhgt = pop_stack();
    ComValue targy = pop_stack();
    ComValue targx = pop_stack();
    ComValue obshgt = pop_stack();
    ComValue obsy = pop_stack();
    ComValue obsx = pop_stack();
    ComValue endaz = pop_stack();
    ComValue startaz = pop_stack();
    ComValue dempath = pop_stack();
    reset_stack();

    if (!dem)
      dem = load_dem((char*)dempath.string_ptr());
    if (set_settings
	(startaz.float_val(),  /* start azimuth */
	 endaz.float_val(), /* end azimuth */
	 obsx.float_val(), /* observer's x */
	 obsy.float_val(), /* observer's y */
	 obshgt.float_val(),  /* observer's height */
	 targx.float_val(), /* target's x */
	 targy.float_val(), /* target's y */
	 targhgt.float_val(),  /* target's height */
	 targwid.float_val(),  /* target's width */
	 range.float_val(), /* range (must be >= minimum) */
	 30, /* vertical accuracy (of DEM?) */
	 0, /* no effect */
	 valtype.int_val(), /* 0 = 2val, 1 = 4val */
	 lostype.int_val() /* 0=viewshed, 1=profile */
	 ) == 0) 
      {
	do_los();
	int nfeats = 0;
	for (int row = prod_height() - 1; row >= 0; row--) {
	  for (int col = 0; col < prod_width(); col++) {
	    unsigned char val = prod_ghandle()
	      [row * prod_width() + col];
	    if (val > 0)
	      nfeats++;
	  }
	}
	if (nfeats == 0) {
	  ComValue result = ComValue::nullval();
	  push_stack(result);
	  return;
	}
	MapFeature** feats = new MapFeature*[nfeats];
	PJ* proj = dem->projection();
	UV data;
	int fi = 0;
	for (int row = prod_height() - 1; row >= 0; row--) {
	  for (int col = 0; col < prod_width(); col++) {
	    unsigned char val = prod_ghandle()
	      [row * prod_width() + col];
	    if (val > 0) {
	      MapFeature* mf = new MapFeature(-1, MapFeature::AreaType);
	      AttributeList* alist = new AttributeList();
	      long int vval = val;
	      AttributeValue* value = new AttributeValue(vval);
	      alist->add_attr("vis", value);
	      mf->attrlist(alist);

	      MapFace* face = new MapFace(mf);
	      float fx[4], fy[4];

	      double lon = prod_ext_left() + 
		col * prod_xres();
	      double lat = prod_ext_bottom() +
		row * prod_yres();
	      double leftlon = lon - prod_xres() / 2.0;
	      double rightlon = lon + prod_xres() / 2.0;
	      double bottomlat = lat - prod_yres() / 2.0;
	      double toplat = lat + prod_yres() / 2.0;
	      
	      data.u = leftlon * DEG_TO_RAD;
	      data.v = bottomlat * DEG_TO_RAD;
	      data = pj_fwd(data, proj);
	      fx[0] = data.u;
	      fy[0] = data.v;
		  
	      data.u = leftlon * DEG_TO_RAD;
	      data.v = toplat * DEG_TO_RAD;
	      data = pj_fwd(data, proj);
	      fx[1] = data.u;
	      fy[1] = data.v;
		  
	      data.u = rightlon * DEG_TO_RAD;
	      data.v = toplat * DEG_TO_RAD;
	      data = pj_fwd(data, proj);
	      fx[2] = data.u;
	      fy[2] = data.v;

	      data.u = rightlon * DEG_TO_RAD;
	      data.v = bottomlat * DEG_TO_RAD;
	      data = pj_fwd(data, proj);
	      fx[3] = data.u;
	      fy[3] = data.v;

	      face->insert_points(4, fx, fy);
	      mf->primitive(face);
	      feats[fi++] = mf;
	    }
	  }
	}
	MapFeatureClass* mapfc = new MapFeatureClass(nfeats, feats);
	MapFClassComp* mfccomp = new MapFClassComp(mapfc);
	for (int i = 0; i < nfeats; i++) {
	  MapFeatureComp* fcomp = new MapFeatureComp(feats[i], mfccomp);
	  //cout << "attribute list: ";
	  //cout << *fcomp->GetAttributeList();
	  //cout << "\n";
	  mfccomp->Append(fcomp);
	}
	AttributeList* vlist = mfccomp->GetAttributeList();
	vlist->add_attr("start_azimuth",
			new AttributeValue(startaz.float_val()));
	vlist->add_attr("end_azimuth",
			new AttributeValue(endaz.float_val()));
	vlist->add_attr("observer_x",
			new AttributeValue(obsx.float_val()));
	vlist->add_attr("observer_y",
			new AttributeValue(obsy.float_val()));
	vlist->add_attr("observer_height",
			new AttributeValue(obshgt.float_val()));
	vlist->add_attr("target_x",
			new AttributeValue(targx.float_val()));
	vlist->add_attr("target_y",
			new AttributeValue(targy.float_val()));
	vlist->add_attr("target_height",
			new AttributeValue(targhgt.float_val()));
	vlist->add_attr("target_width",
			new AttributeValue(targwid.float_val()));
	vlist->add_attr("range",
			new AttributeValue(range.float_val()));
	vlist->add_attr("value_type",
			new AttributeValue(valtype.int_val(), AttributeValue::IntType));
	vlist->add_attr("los_type",
			new AttributeValue(lostype.int_val(), AttributeValue::IntType));
	ComponentView* view = new ComponentView(mfccomp);
	ComValue result;
	result.type(AttributeValue::ObjectType);
	result.obj_ref() = view;
	result.obj_type_ref() = FEATURE_CLASS;
	push_stack(result);
      }
    else {
      ComValue result = ComValue::nullval();
      push_stack(result);
    }
}

/*****************************************************************************/

LosVisUtmFunc::LosVisUtmFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void LosVisUtmFunc::execute() {
  ComValue py = pop_stack();
  ComValue px = pop_stack();
  ComValue losfc = pop_stack();
  reset_stack();
  ComValue result(0, ComValue::IntType);

  ComponentView* view = (ComponentView*)losfc.obj_val();
  MapFClassComp* fccomp = (MapFClassComp*)view->GetSubject();
  float fx = px.float_val();
  float fy = py.float_val();
  FPointObj pnt(fx, fy);
  Iterator i;
  for (fccomp->First(i); !fccomp->Done(i); fccomp->Next(i)) {
    MapFeatureComp* fcomp = (MapFeatureComp*)fccomp->GetComp(i);
    MapFeature* feat = fcomp->feature();
    if (feat->type() == MapFeature::PointType) {
      MapPoint* pt = feat->point_primitive();
      const float* px = pt->xpoints();
      const float* py = pt->ypoints();
      if (px[0] == pnt._x && py[0] == pnt._y) {
	Attribute* attr = fcomp->GetAttributeList()->GetAttr("vis");
	result.int_ref() = attr->Value()->int_val();
	break;
      }
    }
    else if (feat->type() == MapFeature::LineType) {
      MapEdge* edg = feat->edge_primitive();
      float* lx = (float*)edg->xpoints();
      float* ly = (float*)edg->ypoints();
      int n = edg->npts();
      FMultiLineObj ml(lx, ly, n);
      if (ml.Contains(pnt)) {
	Attribute* attr = fcomp->GetAttributeList()->GetAttr("vis");
	result.int_ref() = attr->Value()->int_val();
	break;
      }
    }
    else if (feat->type() == MapFeature::AreaType) {
      MapFace* fac = feat->face_primitive();
      float* fx = (float*)fac->xpoints();
      float* fy = (float*)fac->ypoints();
      int n = fac->npts();
      FFillPolygonObj pol(fx, fy, n);
      if (pol.Contains(pnt)) {
	Attribute* attr = fcomp->GetAttributeList()->GetAttr("vis");
	result.int_ref() = attr->Value()->int_val();
	break;
      }
    }
  }
  push_stack(result);
}

/*****************************************************************************/

DemElevGeoFunc::DemElevGeoFunc(ComTerp* terp) : ComFunc(terp) {
}

void DemElevGeoFunc::execute() {
  ComValue py = pop_stack();
  ComValue px = pop_stack();
  reset_stack();
  double val;
  int good;
  if (dem) {
    if (px.type() == ComValue::FloatType)
      good = dem->value_by_geopos(px.float_val(), py.float_val(), val);
    else if (px.type() == ComValue::DoubleType)
      good = dem->value_by_geopos(px.double_val(), py.double_val(), val);
  }
  ComValue result(val);
  push_stack(result);
}

/*****************************************************************************/

DemElevUtmFunc::DemElevUtmFunc(ComTerp* terp) : ComFunc(terp) {
}

void DemElevUtmFunc::execute() {
  ComValue py = pop_stack();
  ComValue px = pop_stack();
  reset_stack();
  double val;
  int good;
  if (dem) {
    if (px.type() == ComValue::FloatType)
      good = dem->value_by_utmpos(px.float_val(), py.float_val(), val);
    else if (px.type() == ComValue::DoubleType)
      good = dem->value_by_utmpos(px.double_val(), py.double_val(), val);
  }
  ComValue result(val);
  push_stack(result);
}

/*****************************************************************************/

DemTriUtmFunc::DemTriUtmFunc(ComTerp* terp) : ComFunc(terp) {
}

void DemTriUtmFunc::execute() {
  ComValue py = pop_stack();
  ComValue px = pop_stack();
  ComValue dempath = pop_stack();
  reset_stack();
  double x1,y1,z1,x2,y2,z2,x3,y3,z3;
  int good = 0;
  if (!dem)
    dem = load_dem((char*)dempath.string_ptr());
  if (dem) {
    if (px.type() == ComValue::FloatType)
      good = dem->tri3d_by_utmpos(px.float_val(), py.float_val(),
				   x1,y1,z1,x2,y2,z2,x3,y3,z3);
    else if (px.type() == ComValue::DoubleType)
      good = dem->tri3d_by_utmpos(px.double_val(), py.double_val(),
				   x1,y1,z1,x2,y2,z2,x3,y3,z3);
  }
  if (good) {
    MapFeature* mf = new MapFeature(-1, MapFeature::AreaType);
    MapFace* face = new MapFace(mf);
    float fx[3], fy[3];
    fx[0] = x1;
    fx[1] = x2;
    fx[2] = x3;
    fy[0] = y1;
    fy[1] = y2;
    fy[2] = y3;
    face->insert_points(3, fx, fy);
    mf->primitive(face);
    MapFeatureComp* fcomp = new MapFeatureComp(mf);
    AttributeList* alist = fcomp->GetAttributeList();
    alist->add_attr("z1", new AttributeValue(z1));
    alist->add_attr("z2", new AttributeValue(z2));
    alist->add_attr("z3", new AttributeValue(z3));

#ifdef VTK
    float v1[3], v2[3], v3[3], n[3];
    v1[0] = x1;
    v1[1] = y1;
    v1[2] = z1;
    v2[0] = x2;
    v2[1] = y2;
    v2[2] = z2;
    v3[0] = x3;
    v3[1] = y3;
    v3[2] = z3;
    vtkTriangle::ComputeNormal(v1, v2, v3, n);
    alist->add_attr("normalx", new AttributeValue(n[0]));
    alist->add_attr("normaly", new AttributeValue(n[1]));
    alist->add_attr("normalz", new AttributeValue(n[2]));
#endif /* defined(VTK) */
    //alist->add_attr("gs", new AttributeValue(0, AttributeValue::IntType));

    MapFeature** feats = new MapFeature*[1];
    feats[0] = mf;
    MapFeatureClass* mapfc = new MapFeatureClass(1, feats);
    MapFClassComp* mfccomp = new MapFClassComp(mapfc);
    fcomp->SetParent(fcomp, mfccomp);
    mfccomp->Append(fcomp);

    alist->add_attr("proj", new AttributeValue("utm"));
    alist->add_attr("ellps", new AttributeValue("WGS84"));
    double dx1, dy1, dx2, dy2;
    int zone;
    mapfc->extent(dx1, dy1, dx2, dy2, zone);
    AttributeValue* zoneval = new AttributeValue(zone, AttributeValue::IntType);
    alist->add_attr("zone", zoneval);
    AttributeValue* sweval = new AttributeValue(dx1);
    alist->add_attr("sw_e", sweval);
    AttributeValue* swnval = new AttributeValue(dy1);
    alist->add_attr("sw_n", swnval);
    AttributeValue* neeval = new AttributeValue(dx2);
    alist->add_attr("ne_e", neeval);
    AttributeValue* nenval = new AttributeValue(dy2);
    alist->add_attr("ne_n", nenval);

    ComponentView* view = new ComponentView(mfccomp);
    ComValue result;
    result.type(AttributeValue::ObjectType);
    result.obj_ref() = view;
    result.obj_type_ref() = FEATURE_CLASS;
    push_stack(result);
  }
}

/*****************************************************************************/

static int utmzone(float lon) {
  if (lon > -180.0 && lon < -174.0)
    return 1;
  else if (lon > -174.0 && lon < -168.0)
    return 2;
  else if (lon > -168.0 && lon < -162.0)
    return 3;
  else if (lon > -162.0 && lon < -156.0)
    return 4;
  else if (lon > -156.0 && lon < -150.0)
    return 5;
  else if (lon > -150.0 && lon < -144.0)
    return 6;
  else if (lon > -144.0 && lon < -138.0)
    return 7;
  else if (lon > -138.0 && lon < -132.0)
    return 8;
  else if (lon > -132.0 && lon < -126.0)
    return 9;
  else if (lon > -126.0 && lon < -120.0)
    return 10;
  else if (lon > -120.0 && lon < -114.0)
    return 11;
  else if (lon > -114.0 && lon < -108.0)
    return 12;
  else if (lon > -108.0 && lon < -102.0)
    return 13;
  else if (lon > -102.0 && lon < -96.0)
    return 14;
  else if (lon > -96.0 && lon < -90.0)
    return 15;
  else if (lon > -90.0 && lon < -84.0)
    return 16;
  else if (lon > -84.0 && lon < -78.0)
    return 17;
  else if (lon > -78.0 && lon < -72.0)
    return 18;
  else if (lon > -72.0 && lon < -66.0)
    return 19;
  else if (lon > -66.0 && lon < -60.0)
    return 20;
  else if (lon > -60.0 && lon < -54.0)
    return 21;
  else if (lon > -54.0 && lon < -48.0)
    return 22;
  else if (lon > -48.0 && lon < -42.0)
    return 23;
  else if (lon > -42.0 && lon < -36.0)
    return 24;
  else if (lon > -36.0 && lon < -30.0)
    return 25;
  else if (lon > -30.0 && lon < -24.0)
    return 26;
  else if (lon > -24.0 && lon < -18.0)
    return 27;
  else if (lon > -18.0 && lon < -12.0)
    return 28;
  else if (lon > -12.0 && lon < -6.0)
    return 29;
  else if (lon > -6.0 && lon < 0.0)
    return 30;
  else if (lon > 0.0 && lon < 6.0)
    return 31;
  else if (lon > 6.0 && lon < 12.0)
    return 32;
  else if (lon > 12.0 && lon < 18.0)
    return 33;
  else if (lon > 18.0 && lon < 24.0)
    return 34;
  else if (lon > 24.0 && lon < 30.0)
    return 35;
  else if (lon > 30.0 && lon < 36.0)
    return 36;
  else if (lon > 36.0 && lon < 42.0)
    return 37;
  else if (lon > 42.0 && lon < 48.0)
    return 38;
  else if (lon > 48.0 && lon < 54.0)
    return 39;
  else if (lon > 54.0 && lon < 60.0)
    return 40;
  else if (lon > 60.0 && lon < 66.0)
    return 41;
  else if (lon > 66.0 && lon < 72.0)
    return 42;
  else if (lon > 72.0 && lon < 78.0)
    return 43;
  else if (lon > 78.0 && lon < 84.0)
    return 44;
  else if (lon > 84.0 && lon < 90.0)
    return 45;
  else if (lon > 90.0 && lon < 96.0)
    return 46;
  else if (lon > 96.0 && lon < 102.0)
    return 47;
  else if (lon > 102.0 && lon < 108.0)
    return 48;
  else if (lon > 108.0 && lon < 114.0)
    return 49;
  else if (lon > 114.0 && lon < 120.0)
    return 50;
  else if (lon > 120.0 && lon < 126.0)
    return 51;
  else if (lon > 126.0 && lon < 132.0)
    return 52;
  else if (lon > 132.0 && lon < 138.0)
    return 53;
  else if (lon > 138.0 && lon < 144.0)
    return 54;
  else if (lon > 144.0 && lon < 150.0)
    return 55;
  else if (lon > 150.0 && lon < 156.0)
    return 56;
  else if (lon > 156.0 && lon < 162.0)
    return 57;
  else if (lon > 162.0 && lon < 168.0)
    return 58;
  else if (lon > 168.0 && lon < 174.0)
    return 59;
  else if (lon > 174.0 && lon < 180.0)
    return 60;
  else
    return 0;
}

inline double dmin(double a, double b) { return a < b ? a : b; } \
inline double dmax(double a, double b) { return a > b ? a : b; }

DemGridUtmFunc::DemGridUtmFunc(ComTerp* terp) : ComFunc(terp) {
}

void DemGridUtmFunc::execute() {
  ComValue py1 = pop_stack();
  ComValue px1 = pop_stack();
  ComValue py0 = pop_stack();
  ComValue px0 = pop_stack();
  ComValue dempath = pop_stack();
  reset_stack();
  double *tx, *ty, *tz;
  int nt;
  int good = 0;
  if (!dem)
    dem = load_dem((char*)dempath.string_ptr());
  if (dem) {
    if (px0.type() == ComValue::FloatType)
      good = dem->trimesh_by_utmpos(px0.float_val(), py0.float_val(),
				     px1.float_val(), py1.float_val(),
				     nt, tx, ty, tz);
    else if (px0.type() == ComValue::DoubleType)
      good = dem->trimesh_by_utmpos(px0.double_val(), py0.double_val(),
				     px1.double_val(), py1.double_val(),
				     nt, tx, ty, tz);
  }
  if (good) {
    MapFeature** feats = new MapFeature*[nt/3];
    MapFeatureClass* mapfc = new MapFeatureClass(nt/3, feats);
    MapFClassComp* mfccomp = new MapFClassComp(mapfc);
    int fi = 0;

    double xmin = MAXDOUBLE;
    double ymin = MAXDOUBLE;
    double xmax = MINDOUBLE;
    double ymax = MINDOUBLE;
    for (int ti = 0; ti < nt; ti += 3) {
      xmin = dmin(xmin, tx[ti]);
      xmin = dmin(xmin, tx[ti+1]);
      xmin = dmin(xmin, tx[ti+2]);
      ymin = dmin(ymin, ty[ti]);
      ymin = dmin(ymin, ty[ti+1]);
      ymin = dmin(ymin, ty[ti+2]);
      xmax = dmax(xmax, tx[ti]);
      xmax = dmax(xmax, tx[ti+1]);
      xmax = dmax(xmax, tx[ti+2]);
      ymax = dmax(ymax, ty[ti]);
      ymax = dmax(ymax, ty[ti+1]);
      ymax = dmax(ymax, ty[ti+2]);

      MapFeature* mf = new MapFeature(-1, MapFeature::AreaType);
      MapFace* face = new MapFace(mf);
      float fx[3], fy[3];
      fx[0] = tx[ti];   //x1;
      fx[1] = tx[ti+1]; //x2;
      fx[2] = tx[ti+2]; //x3;
      fy[0] = ty[ti];   //y1;
      fy[1] = ty[ti+1]; //y2;
      fy[2] = ty[ti+2]; //y3;
      face->insert_points(3, fx, fy);
      mf->primitive(face);
      MapFeatureComp* fcomp = new MapFeatureComp(mf);
      AttributeList* alist = fcomp->GetAttributeList();
      alist->add_attr("z1", new AttributeValue(tz[ti]/*z1*/));
      alist->add_attr("z2", new AttributeValue(tz[ti+1]/*z2*/));
      alist->add_attr("z3", new AttributeValue(tz[ti+2]/*z3*/));

#ifdef VTK
      float v1[3], v2[3], v3[3], n[3];
      v1[0] = tx[ti]; //x1;
      v1[1] = ty[ti]; //y1;
      v1[2] = tz[ti]; //z1;
      v2[0] = tx[ti+1]; //x2;
      v2[1] = ty[ti+1]; //y2;
      v2[2] = tz[ti+1]; //z2;
      v3[0] = tx[ti+2]; //x3;
      v3[1] = ty[ti+2]; //y3;
      v3[2] = tz[ti+2]; //z3;
      vtkTriangle::ComputeNormal(v1, v2, v3, n);
      alist->add_attr("normalx", new AttributeValue(n[0]));
      alist->add_attr("normaly", new AttributeValue(n[1]));
      alist->add_attr("normalz", new AttributeValue(n[2]));
#endif
      //alist->add_attr("gs", new AttributeValue(0, AttributeValue::IntType));

      feats[fi++] = mf;
      fcomp->SetParent(fcomp, mfccomp);
      mfccomp->Append(fcomp);
    }

    AttributeList* alist = mfccomp->GetAttributeList();
    alist->add_attr("proj", new AttributeValue("utm"));
    alist->add_attr("ellps", new AttributeValue("WGS84"));
    int zone = utmzone(xmin);
    mapfc->set_extent(xmin, ymin, xmax, ymax, zone);
    if (zone) {
      AttributeValue* zoneval = new AttributeValue(zone, AttributeValue::IntType);
      alist->add_attr("zone", zoneval);
    }
    AttributeValue* sweval = new AttributeValue(xmin);
    alist->add_attr("sw_e", sweval);
    AttributeValue* swnval = new AttributeValue(ymin);
    alist->add_attr("sw_n", swnval);
    AttributeValue* neeval = new AttributeValue(xmax);
    alist->add_attr("ne_e", neeval);
    AttributeValue* nenval = new AttributeValue(ymax);
    alist->add_attr("ne_n", nenval);

    ComponentView* view = new ComponentView(mfccomp);
    ComValue result;
    result.type(AttributeValue::ObjectType);
    result.obj_ref() = view;
    result.obj_type_ref() = FEATURE_CLASS;
    push_stack(result);
  }
}
/*****************************************************************************/

DemXYZFunc::DemXYZFunc(ComTerp* comterp) : ComFunc(comterp) {
  _basename_symid = symbol_add("basename");
  _swe_symid = symbol_add("sw_e");
  _swn_symid = symbol_add("sw_n");
  _tsize_symid = symbol_add("tsize");
  _samplerate_symid = symbol_add("samplerate");
  _zone_symid = symbol_add("zone");
}

void DemXYZFunc::execute() {
  ComValue basename(stack_key(_basename_symid));
  ComValue swe(stack_key(_swe_symid));
  ComValue swn(stack_key(_swn_symid));
  ComValue tsize(stack_key(_tsize_symid));
  ComValue samplerate(stack_key(_samplerate_symid));
  ComValue zone(stack_key(_zone_symid));

  int na = nargsfixed();
  long xmin, ymin, xmax, ymax;
  if (na > 0) {
    for (int i = 0; i < na; i++) {
      ComValue dmpath(stack_arg(i));
      if (dmpath.type() != ComValue::StringType) {
	reset_stack();
	push_stack(ComValue::nullval());
	return;
      }
    }
    UsgsDem* dems[na];
    for (int i = 0; i < na; i++) {
      dems[i] = new UsgsDem();
    }
    FBoxObj geobox;
    for (int i = 0; i < na; i++) {
      ComValue dmpath(stack_arg(i));
      dems[i]->read((char*)dmpath.string_ptr());
      if (dems[i]->rows()==0 || dems[i]->cols()==0) {
	dems[i] = nil;
      }
      else {
	if (i == 0) {
	  if (dems[i]->reference_system() == UsgsDem::GEOGRAPHIC) {
	    geobox._left = dems[i]->swe();
	    geobox._bottom = dems[i]->swn();
	    geobox._right = dems[i]->nee();
	    geobox._top = dems[i]->nen();
	  }
	  else if (dems[i]->reference_system() == UsgsDem::UTM) {
	    dems[i]->utm_bbox(xmin, ymin, xmax, ymax);
	    geobox._left = xmin;
	    geobox._bottom = ymin;
	    geobox._right = xmax;
	    geobox._top = ymax;
	  }
	}
	else {
	  if (dems[i]->reference_system() == UsgsDem::GEOGRAPHIC) {
	    FBoxObj dbox(dems[i]->swe(), dems[i]->swn(), dems[i]->nee(),
			 dems[i]->nen());
	    geobox = geobox + dbox;
	  }
	  else if (dems[i]->reference_system() == UsgsDem::UTM) {
	    dems[i]->utm_bbox(xmin, ymin, xmax, ymax);
	    FBoxObj dbox(xmin, ymin, xmax, ymax);
	    geobox = geobox + dbox;
	  }
	}
      }
    }

    const char* bnm = basename.unknown() ? "demxyz" : basename.string_ptr();
    double swx = swe.unknown() ? geobox._left : swe.double_val();
    double swy = swn.unknown() ? geobox._bottom : swn.double_val();
    int ts = tsize.unknown() ? 512 : tsize.int_val();
    int srate = samplerate.unknown() ? 30 : samplerate.int_val();
    int uz = zone.int_val();

    char* parms[4];
    char zstr[10];
    sprintf(zstr, "zone=%d", uz);
    parms[0] = "proj=utm";
    parms[1] = "ellps=WGS84";
    parms[2] = zstr;
    parms[3] = "no_defs";
    PJ* proj = pj_init(4, parms);

    int xtiles = (Math::round(geobox._right - geobox._left) /
		  (ts * srate)) + 1;
    int ytiles = (Math::round(geobox._top - geobox._bottom) /
		  (ts * srate)) + 1;
    if (1 || dems[0]->reference_system() == UsgsDem::GEOGRAPHIC) {
    for (int yt = 0; yt < ytiles; yt++) {
      for (int xt = 0; xt < xtiles; xt++) {
	char fname[1024] = "";
	sprintf(fname, "%s.%d.%d.xyz", bnm ? bnm : "dem", xt, yt);
	FILE* fp = fopen(fname, "w");

	for (int yi = ts; yi >= 0; yi--)
	  for (int xi = 0; xi < ts; xi++) {
	    double ux = swx + (xt*ts*srate) + (xi * srate);
	    double uy = swy + (yt*ts*srate) + (yi * srate);
	    if (xi == 0 && yi == 0)
	      fprintf(stderr, "%s: xmin %0.6f ymin %0.6f\n", fname, ux, uy);
	    UV data;
	    data.u = ux;
	    data.v = uy;
	    data = pj_inv(data, proj);
	    double maxval = 0.0;
	    double val = 0.0;
	    for (int i = 0; i < na; i++) {
	      if ((dems[i]->horizontal_units() == arcseconds &&
		   dems[i]->geowithin(data.u*RAD_TO_DEG, data.v*RAD_TO_DEG))
		  ||
		  (dems[i]->horizontal_units() == meters /*&&
		   dems[i]->geowithin(ux, uy)*/)) {
		dems[i]->value_by_geopos(data.u*RAD_TO_DEG, 
					  data.v*RAD_TO_DEG,
					  val);
		if (isinf(val) || isnan(val) || val < 0.0)
		  val = 0.0;
		if (dems[i]->elevation_units() == UsgsDem::FEET) {
		  val /= 3.2808399;
		}
		//break;
		maxval = max(maxval, val);
	      }
	    }
	    fprintf(fp, "%0.1f %0.1f %0.6f\n", ux, uy, maxval);
	  }
	fclose(fp);
      
      }
    }
    for (int i = 0; i < na; i++) {
      delete dems[i];
    }
    }

    else {
      UsgsDem* mdem = new UsgsDem();
      mdem->reference_system(dems[0]->reference_system());
      mdem->ground_units(dems[0]->ground_units());
      mdem->elevation_units(dems[0]->elevation_units());
      mdem->horizontal_units(dems[0]->horizontal_units());
      mdem->vertical_units(dems[0]->vertical_units());
      mdem->vertical_datum(dems[0]->vertical_datum());
      mdem->horizontal_datum(dems[0]->horizontal_datum());
      mdem->projection(dems[0]->projection());
      double res = dems[0]->x_res();
      mdem->res(res, res);
      mdem->set_utm_bbox(long(geobox._left), long(geobox._bottom),
		     long(geobox._right), long(geobox._top));
      int nrows = int((geobox._top - geobox._bottom) / res);
      int ncols = int((geobox._right - geobox._left) / res);
      mdem->allocate(nrows, ncols);
      for (int i = 0; i < na; i++) {
	for (int ci = 0; ci < dems[i]->cols(); ci++) {
	  for (int ri = dems[i]->start_row(ci);
	       ri <= dems[i]->end_row(ci);
	       ri++)
	    {
	      unsigned short v = dems[i]->value(ri, ci);
	      double px, py;
	      dems[i]->position(ri, ci, px, py);
	      if (dems[i]->horizontal_units() == feet)
		v = Math::round(v /  3.2808399);
	      mdem->value_at_utmpos(long(px), long(py), v);
	    }
	}
      }

    for (int yt = 0; yt < ytiles; yt++) {
      for (int xt = 0; xt < xtiles; xt++) {
	char fname[1024] = "";
	sprintf(fname, "%s.%d.%d.xyz", bnm ? bnm : "dem", xt, yt);
	FILE* fp = fopen(fname, "w");

	for (int yi = ts; yi >= 0; yi--)
	  for (int xi = 0; xi < ts; xi++) {
	    double ux = geobox._left + (xt*ts*srate) + (xi * srate);
	    double uy = geobox._bottom + (yt*ts*srate) + (yi * srate);
	    if (xi == 0 && yi == 0)
	      fprintf(stderr, "%s: xmin %0.6f ymin %0.6f\n", fname, ux, uy);
	    UV data;
	    data.u = ux;
	    data.v = uy;
	    data = pj_inv(data, proj);
	    double maxval = 0.0;
	    double val = 0.0;
	      mdem->value_by_utmpos(ux, uy, val);
		if (isinf(val) || isnan(val) || val < 0.0)
		  val = 0.0;
		if (mdem->elevation_units() == UsgsDem::FEET)
		  val /= 3.2808399;
	    fprintf(fp, "%0.1f %0.1f %0.6f\n", ux, uy, val);
	  }
	fclose(fp);
      
      }
    }
    for (int i = 0; i < na; i++) {
      delete dems[i];
    }
      
    }
    reset_stack();
  }
}
