/*
 * Copyright 1994, 1995, 1998 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Map/mapprim.h>
#include <Vpf/vpfcov.h>
#include <Vpf/vpffeature.h>
#include <Vpf/vpftiledfc.h>
#include <Vpf/vpflib.h>
#include <OS/math.h>

#include <iostream.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

extern "C" {
#include <malloc.h>
#include <sys/time.h>
#include <sys/types.h>
}
#include <stdio.h>

/*****************************************************************************/

VpfTiledFeatureClass::VpfTiledFeatureClass(VpfCoverage* vc, const char* name,
					   const char* tilename)
: VpfFeatureClass(vc, name)
{
  _tilename = tilename ? strdup(tilename) : nil;
  if (!_tilename) {
    VpfFeatureClass* tileref = coverage()->library()->coverage("tileref")->
	feature_class("tileref");
    _ntiles = tileref->nfeats();
  }
  else
    _ntiles = 1;

    tilestart = new int[_ntiles+3];
    tileend = new int[_ntiles+3];
    for (int i = 0; i < _ntiles+3; i++)
	tilestart[i] = tileend[i] = 0;
    _tfeatures = new VpfFeature**[_ntiles+3];
    for (int i = 0; i < _ntiles+3; i++)
	_tfeatures[i] = nil;
    tiled(true);
}

VpfTiledFeatureClass::~VpfTiledFeatureClass() {
    for (int i = 0; i < _ntiles+3; i++) {
	if (_tfeatures[i]) {
	    for (int j = 0; j < nfeats(i); j++)
		delete (_tfeatures[i])[j];
	    delete [] _tfeatures[i];
	}
    }
    delete [] _tfeatures;
    delete [] tilestart;
    delete [] tileend;
}

/* this will allocate the array but not create features */

MapFeature** VpfTiledFeatureClass::make_features(int &nfeats) const {
    vpf_table_type feattable;
    char featpath[255];
    strcpy(featpath, _tabl);

    feattable = vpf_open_table(featpath,disk,"rb",NULL);
    if (!feattable.fp) return false;
    nfeats = feattable.nrows;
    vpf_close_table( &feattable );
    return nil;
}

void fixpath(char* path) {
    int i = 0;
    while (path[i]) {
	if (path[i]=='\\')
	    path[i]='/';
	else {
	    int c = path[i];
	    c = tolower(path[i]);
	    path[i] = c;
	}
	i++;
    }
}

static char* strupr(char* str) {
  char* s = str;
  while (*s)
    *s++ = toupper(*s);
  return str;
}

static char *rjust( char *str )
{
   register int  len,i;

   len = strlen(str);
   i = len - 1;
   while ((i>0) && ((str[i]==0) || (str[i]==' '))) i--;
   if (i < (len-1)) str[i+1] = '\0';
   for (i=0;i<strlen(str);i++) if (str[i]=='\n') str[i] = '\0';
   return str;
}

void VpfTiledFeatureClass::init_tile(int tileid) {
    char* tilename;
    long int count;
    if (!_tilename) {
      VpfFeatureClass* tileref = coverage()->library()->coverage("tileref")->
	feature_class("tileref");
      VpfFeature* tile = tileref->feature(tileid-1);
      tilename = (char*)tile->named_row_element("TILE_NAME", (void*)NULL, &count);
    }
    else
      tilename = strdup(_tilename);

    vpf_table_type table;
    char tablepath[255];
    fixpath(tilename);
    rjust(tilename);
    strcpy(tablepath, _coverage->path());
    strcat(tablepath, "/");
    strcat(tablepath, tilename);
    strcat(tablepath, &_ptable[strlen(_ptable)-5]);

    struct stat statbuf;
    if (stat(tablepath, &statbuf) != 0 &&
	(tablepath[strlen(tablepath)-1]='\0', stat(tablepath, &statbuf) != 0)
    ) {
	tilestart[tileid] = 0;
	tileend[tileid] = -1;
    }
    else {
	table = vpf_open_table(tablepath, disk, "rb", NULL);
	row_type row;
	char colname[20];
	strcpy(colname, strrchr(_tabl, '/')+1);
	strupr(colname);
	strcat(colname, "_ID");
	long int colpos = table_pos(colname, table);
	long int n;
	row = get_row(1, table);
	get_table_element(colpos,row,table,&tilestart[tileid],&n);
	free_row(row,table);
	row = get_row(table.nrows, table);
	get_table_element(colpos,row,table,&tileend[tileid],&n);
	free_row(row,table);
	vpf_close_table(&table);
    }
    free(tilename);
}

int VpfTiledFeatureClass::nfeats(int tileid) {
    if (tilestart[tileid] == 0)
	init_tile(tileid);
    return tileend[tileid]-tilestart[tileid]+1;
}

VpfFeature* VpfTiledFeatureClass::feature(int tileid, int fnum) {
    if (_tfeatures[tileid] == nil)
	make_tile_features(tileid);
    return (_tfeatures[tileid])[fnum];
}

void VpfTiledFeatureClass::make_tile_features(int tileid) {
    char* tilename;
    long int count;
    if (!_tilename) {
      VpfFeatureClass* tileref = coverage()->library()->coverage("tileref")->
	feature_class("tileref");
      VpfFeature* tile = tileref->feature(tileid-1);
      tilename = (char*)tile->named_row_element("TILE_NAME", (void*)NULL, &count);
      fixpath(tilename);
      rjust(tilename);
    }
    else
      tilename = strdup(_tilename);
    if (_fctype == AREA) {
	vpf_table_type feattable, facetable, ringtable, edgetable, ebrtable;
	long int   i,j,k,n;
	char       *buf,  ch, date[40];
	row_type   featrow, facerow, ringrow, edgerow, ebrrow;
	char featpath[255], facepath[255], ringpath[255], edgepath[255];
	char ebrpath[255];
	boolean notype = false;
	
	strcpy(featpath, _tabl);
	strcpy(facepath, _coverage->path());
	strcat(facepath, "/");
	strcat(facepath, tilename);
	strcat(facepath, "/fac.");
	strcpy(ringpath, _coverage->path());
	strcpy(edgepath, ringpath);
	strcpy(ebrpath, ringpath);
	strcat(ringpath, "/");
	strcat(ringpath, tilename);
	strcat(ringpath, "/rng.");
	strcat(edgepath, "/");
	strcat(edgepath, tilename);
	strcat(edgepath, "/edg.");
	strcat(ebrpath, "/");
	strcat(ebrpath, tilename);
	strcat(ebrpath, "/ebr.");
	
	feattable = vpf_open_table(featpath,disk,"rb",NULL);
	if (!feattable.fp) return;
	long int feat_id_pos = table_pos("ID", feattable);
	if (strcmp(name(),"tileref")==0 || strcmp(name(),"clarea")==0)
	    notype = true;
	long int feat_type_pos;
	if (!notype) {
	    char tname[20];
	    const char* fcname = name();
	    tname[0] = toupper(fcname[0]);
	    tname[1] = toupper(fcname[1]);
	    tname[2] = '\0';
	    strcat(tname, "PYTYPE");
	    feat_type_pos = table_pos(tname, feattable);
	    if (feat_type_pos == -1)
	      notype = true;
	}
	long int fcode_pos;
	fcode_pos = table_pos("f_code", feattable);
	facetable = vpf_open_table(facepath,disk,"rb",NULL);
	if (!facetable.fp) return;
	long face_id_pos = table_pos("ID", facetable);
	long face_ringptr_pos = table_pos("RING_PTR", facetable);
	char colname[20];
	strcpy(colname, strrchr(_tabl, '/')+1);
	strupr(colname);
	strcat(colname, "_ID");
	long int face_feat_pos = table_pos(colname, facetable);
	ringtable = vpf_open_table(ringpath,disk,"rb",NULL);
	if (!ringtable.fp) return;
	long ring_id_pos = table_pos("ID", ringtable);
	long ring_faceid_pos = table_pos("FACE_ID", ringtable);
	long ring_startedge_pos = table_pos("START_EDGE", ringtable);
	edgetable = vpf_open_table(edgepath,disk,"rb",NULL);
	if (!edgetable.fp) return;
	long edge_id_pos = table_pos("ID", edgetable);
	long edge_rface_pos = table_pos("RIGHT_FACE", edgetable);
	long edge_lface_pos = table_pos("LEFT_FACE", edgetable);
	long edge_right_pos = table_pos("RIGHT_EDGE", edgetable);
	long edge_left_pos = table_pos("LEFT_EDGE", edgetable);
	long edge_coord_pos = table_pos("COORDINATES", edgetable);
	
	_tfeatures[tileid] = new VpfFeature*[facetable.nrows];
	for (i=1;i<=facetable.nrows;i++) {
	    facerow = get_row(i,facetable);
	    long int faceid, featid, feattype;
	    get_table_element(face_id_pos,facerow,facetable,&faceid,&n);
	    get_table_element(face_feat_pos,facerow,facetable,&featid,&n);
	    if (featid > 0) {
	      featrow = get_row(featid,feattable);
	      if (!notype) {
		get_table_element(feat_type_pos,featrow,feattable,&feattype,&n);
		(_tfeatures[tileid])[i-1] = outline_face(featrow,featid,faceid,facetable,ringtable,edgetable,
					     feattype,0);
	      }
	      else
		(_tfeatures[tileid])[i-1] = outline_face(featrow,featid,faceid,facetable,ringtable,edgetable,
							 0,0);
	    }
	    free_row(facerow,facetable);
	}
	vpf_close_table( &feattable );
	vpf_close_table( &facetable );
	vpf_close_table( &ringtable );
	vpf_close_table( &edgetable );

	ebrtable = vpf_open_table(ebrpath,disk,"rb",NULL);
	if (!ebrtable.fp) return;
	long ebr_xmin_pos = table_pos("XMIN", ebrtable);
	long ebr_ymin_pos = table_pos("YMIN", ebrtable);
	long ebr_xmax_pos = table_pos("XMAX", ebrtable);
	long ebr_ymax_pos = table_pos("YMAX", ebrtable);
	for (i=1;i<=ebrtable.nrows;i++) {
	  float fx1, fy1, fx2, fy2;
	  ebrrow = get_row(i,ebrtable);
	  get_table_element(ebr_xmin_pos,ebrrow,ebrtable,&fx1,&n);
	  get_table_element(ebr_ymin_pos,ebrrow,ebrtable,&fy1,&n);
	  get_table_element(ebr_xmax_pos,ebrrow,ebrtable,&fx2,&n);
	  get_table_element(ebr_ymax_pos,ebrrow,ebrtable,&fy2,&n);
	  if (i==1) {
	    _x1 = fx1; _y1 = fy1; _x2 = fx2; _y2 = fy2;
	  }
	  else {
	    _x1 = Math::min(_x1, fx1);
	    _y1 = Math::min(_y1, fy1);
	    _x2 = Math::max(_x2, fx2);
	    _y2 = Math::max(_y2, fy2);
	  }
	  free_row(ebrrow, ebrtable);
	}
	vpf_close_table( &ebrtable );
    }
    else if (_fctype == LINE) {
	vpf_table_type feattable, edgetable, ebrtable;
	long int   i,k,n;
	char       *buf,  ch, date[40];
	row_type   featrow, edgerow, ebrrow;
	char featpath[255], edgepath[255], ebrpath[255];
	edge_rec_type edge_rec;
	boolean notype = false;
	
	strcpy(featpath, _tabl);
	strcpy(edgepath, _coverage->path());
	strcat(edgepath, "/");
	strcat(edgepath, tilename);
	strcpy(ebrpath, edgepath);
	strcat(edgepath, "/edg.");
	strcat(ebrpath, "/ebr.");
	
	feattable = vpf_open_table(featpath,disk,"rb",NULL);
	if (!feattable.fp) return;
	long int feat_id_pos = table_pos("ID", feattable);
	if (strcmp(name(),"clline")==0)
	    notype = true;
	long int feat_type_pos;
	if (!notype) {
	    char tname[20];
	    const char* fcname = name();
	    tname[0] = toupper(fcname[0]);
	    tname[1] = toupper(fcname[1]);
	    tname[2] = '\0';
	    strcat(tname, "LNTYPE");
	    feat_type_pos = table_pos(tname, feattable);
	    if (feat_type_pos == -1)
	      notype = true;
	}
	edgetable = vpf_open_table(edgepath,disk,"rb",NULL);
	if (!edgetable.fp) return;
	long edge_id_pos = table_pos("ID", edgetable);
	long edge_rface_pos = table_pos("RIGHT_FACE", edgetable);
	long edge_lface_pos = table_pos("LEFT_FACE", edgetable);
	long edge_right_pos = table_pos("RIGHT_EDGE", edgetable);
	long edge_left_pos = table_pos("LEFT_EDGE", edgetable);
	long edge_coord_pos = table_pos("COORDINATES", edgetable);
	char colname[20];
	strcpy(colname, strrchr(_tabl, '/')+1);
	strupr(colname);
	strcat(colname, "_ID");
	long int edge_feat_pos = table_pos(colname, edgetable);

	_tfeatures[tileid] = new VpfFeature*[edgetable.nrows];
	for (i=1;i<=edgetable.nrows;i++) {
	    edgerow = get_row(i,edgetable);
	    long int featid, feattype, featedge, edgeid;
	    get_table_element(edge_id_pos,edgerow,edgetable,&edgeid,&n);
	    get_table_element(edge_feat_pos,edgerow,edgetable,&featid,&n);
	    featrow = get_row(featid,feattable);
	    if (!notype)
		get_table_element(feat_type_pos,featrow,feattable,&feattype,&n);
	    else
		feattype = 0;
	    edge_rec = read_edge(edgeid, edgetable);
	    int n = edge_rec.npts;
	    float* x = new float[n];
	    float* y = new float[n];
	    coordinate_type coord;
	    for (int j = 0; j < n; j++) {
		coord = get_edge_coordinate(j,&edge_rec);
		x[j] = coord.x;
		y[j] = coord.y;
	    }
	    MapEdge* me = new MapEdge;
	    me->insert_pointers(n, x, y, nil, true);
	    (_tfeatures[tileid])[i-1] = new VpfFeature((VpfFeatureClass*)this, featid, MapFeature::LineType, 
					    feattype, featrow);
	    (_tfeatures[tileid])[i-1]->edge_primitive(me);
	    free_row(edgerow,edgetable);
	}
	vpf_close_table( &feattable );
	vpf_close_table( &edgetable );

	ebrtable = vpf_open_table(ebrpath,disk,"rb",NULL);
	if (!ebrtable.fp) return;
	long ebr_xmin_pos = table_pos("XMIN", ebrtable);
	long ebr_ymin_pos = table_pos("YMIN", ebrtable);
	long ebr_xmax_pos = table_pos("XMAX", ebrtable);
	long ebr_ymax_pos = table_pos("YMAX", ebrtable);
	for (i=1;i<=ebrtable.nrows;i++) {
	  float fx1, fy1, fx2, fy2;
	  ebrrow = get_row(i,ebrtable);
	  get_table_element(ebr_xmin_pos,ebrrow,ebrtable,&fx1,&n);
	  get_table_element(ebr_ymin_pos,ebrrow,ebrtable,&fy1,&n);
	  get_table_element(ebr_xmax_pos,ebrrow,ebrtable,&fx2,&n);
	  get_table_element(ebr_ymax_pos,ebrrow,ebrtable,&fy2,&n);
	  if (i==1) {
	    _x1 = fx1; _y1 = fy1; _x2 = fx2; _y2 = fy2;
	  }
	  else {
	    _x1 = Math::min(_x1, fx1);
	    _y1 = Math::min(_y1, fy1);
	    _x2 = Math::max(_x2, fx2);
	    _y2 = Math::max(_y2, fy2);
	  }
	  free_row(ebrrow, ebrtable);
	}
	vpf_close_table( &ebrtable );
    }
    else if (_fctype == POINT) {
	vpf_table_type feattable, nodetable;
	long int   i,j,k,n;
	char       *buf,  ch, date[40];
	row_type   featrow, noderow;
	char featpath[255], nodepath[255];
	point_rec_type node_rec;
	boolean notype = false;
	
	strcpy(featpath, _tabl);
	strcpy(nodepath, _coverage->path());
	strcat(nodepath, "/");
	strcat(nodepath, tilename);
	strcat(nodepath, "/end.");
	
	feattable = vpf_open_table(featpath,disk,"rb",NULL);
	if (!feattable.fp) return;
	long int feat_id_pos = table_pos("ID", feattable);
	if (strcmp(name(),"clpoint")==0 || strcmp(name(),"hspoint")==0)
	    notype = true;
	long int feat_type_pos;
	if (!notype) {
	    char tname[20];
	    const char* fcname = name();
	    tname[0] = toupper(fcname[0]);
	    tname[1] = toupper(fcname[1]);
	    tname[2] = '\0';
	    strcat(tname, "PTTYPE");
	    feat_type_pos = table_pos(tname, feattable);
	    if (feat_type_pos == -1)
	      notype = true;
	}
	nodetable = vpf_open_table(nodepath,disk,"rb",NULL);
	if (!nodetable.fp) return;
	long node_id_pos = table_pos("ID", nodetable);
	long node_coord_pos = table_pos("COORDINATE", nodetable);
	char colname[20];
	strcpy(colname, strrchr(_tabl, '/')+1);
	strupr(colname);
	strcat(colname, "_ID");
	long int node_feat_pos = table_pos(colname, nodetable);
	
	_tfeatures[tileid] = new VpfFeature*[nodetable.nrows];
	for (i=1;i<=nodetable.nrows;i++) {
	    noderow = get_row(i,nodetable);
	    long int featid, feattype, featnode, nodeid;
	    get_table_element(node_id_pos,noderow,nodetable,&nodeid,&n);
	    get_table_element(node_feat_pos,noderow,nodetable,&featid,&n);
	    featrow = get_row(featid,feattable);
	    if (!notype)
		get_table_element(feat_type_pos,featrow,feattable,&feattype,&n);
	    else
		feattype = 0;
	    node_rec = read_point(nodeid, nodetable);
	    float dx = node_rec.x;
	    float dy = node_rec.y;
	    if (i == 1) {
	      _x1 = dx; _y1 = dy; _x2 = dx; _y2 = dy;
	    }
	    else {
	      _x1 = Math::min(_x1, dx);
	      _y1 = Math::min(_y1, dy);
	      _x2 = Math::max(_x2, dx);
	      _y2 = Math::max(_y2, dy);
	    }
	    MapPoint* mp = new MapPoint();
	    mp->insert_points(1, &dx, &dy);
	    (_tfeatures[tileid])[i-1] = new VpfFeature((VpfFeatureClass*)this, featid, MapFeature::PointType, 
					   feattype, featrow);
	    (_tfeatures[tileid])[i-1]->point_primitive(mp);
	}
	vpf_close_table( &feattable );
	vpf_close_table( &nodetable );
    }
    else if (_fctype == ANNO) {
	vpf_table_type feattable, texttable;
	long int   i,j,k,n;
	char       *buf,  ch, date[40];
	row_type   featrow, textrow;
	char featpath[255], textpath[255];
	anno_rec_type anno_rec;
	boolean notype = false;
	
	strcpy(featpath, _tabl);
	strcpy(textpath, _coverage->path());
	strcat(textpath, "/");
	strcat(textpath, tilename);
	strcat(textpath, "/txt.");
	
	feattable = vpf_open_table(featpath,disk,"rb",NULL);
	if (!feattable.fp) return;
	long int feat_id_pos = table_pos("ID", feattable);
	notype = true;
	texttable = vpf_open_table(textpath,disk,"rb",NULL);
	if (!texttable.fp) return;
	long text_id_pos = table_pos("ID", texttable);
	long text_coord_pos = table_pos("SHAPE_LINE", texttable);
	char colname[20];
	strcpy(colname, strrchr(_tabl, '/')+1);
	strupr(colname);
	strcat(colname, "_ID");
	long int text_feat_pos = table_pos(colname, texttable);

	_tfeatures[tileid] = new VpfFeature*[texttable.nrows];
	for (i=1;i<=texttable.nrows;i++) {
	    textrow = get_row(i,texttable);
	    long int featid, feattype, feattext, textid;
	    get_table_element(text_id_pos,textrow,texttable,&textid,&n);
	    get_table_element(text_feat_pos,textrow,texttable,&featid,&n);
	    featrow = get_row(featid,feattable);
	    anno_rec = read_anno(textid, texttable);
	    float dx = anno_rec.x;
	    float dy = anno_rec.y;
	    if (i == 1) {
	      _x1 = dx; _y1 = dy; _x2 = dx; _y2 = dy;
	    }
	    else {
	      _x1 = Math::min(_x1, dx);
	      _y1 = Math::min(_y1, dy);
	      _x2 = Math::max(_x2, dx);
	      _y2 = Math::max(_y2, dy);
	    }
	    MapText* mt = new MapText();
	    mt->text(anno_rec.text);
	    mt->insert_points(1, &dx, &dy);
	    (_tfeatures[tileid])[i-1] = new VpfFeature((VpfFeatureClass*)this, featid, 
						       MapFeature::TextType, 
						       0, featrow);
	    (_tfeatures[tileid])[i-1]->text_primitive(mt);
	}
	vpf_close_table( &feattable );
	vpf_close_table( &texttable );
    }
}
